import importlib


# Constants; because they make the code look nice.
MODULE_PATH_SEP = '.'
MODULE_INIT_FILE = '__init__.py'
NAME_ATTRIBUTE = '__name__'
PATH_ATTRUBITE = '__path__'


def import_module(module_name):
    """
    A nice namespace alias for users.
    """
    return importlib.import_module(module_name)



import os
import shutil
import tempfile
import unittest
import pynsive


INIT_PY = """
from .test_classes import *
"""

TEST_CLASSES_PY = """
SUCCESS = True

class PynsiveTestingClass(object):
    pass


class OtherPynsiveTestingClass(PynsiveTestingClass):
    pass
"""


class WhenLoading(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        """
        This crazy setUp method for the following unit tests creates a
        temporary plugin directory and then drops a Python module and related
        testing code into it. The method then spins up a plugin context via
        a pynsive.PluginManager instance.
        """

        self.directory = tempfile.mkdtemp()
        self.module_path = os.path.join(self.directory, 'pynsive_test')
        os.mkdir(self.module_path)
        with open(os.path.join(self.module_path, '__init__.py'), 'w') as f:
            f.write(INIT_PY)
        with open(os.path.join(self.module_path, 'test_classes.py'), 'w') as f:
            f.write(TEST_CLASSES_PY)
        self.plugin_manager = pynsive.PluginManager()
        self.plugin_manager.plug_into(self.directory)

    @classmethod
    def tearDownClass(self):
        self.plugin_manager.destroy()
        if self.directory:
            shutil.rmtree(self.directory)

    def test_plugging_into_directory(self):
        """
        When plugging into a directory using a PluginManager, the manager
        will make the new directory available for search when importing
        modules. These modules must be available for import via the
        import_module function provided by pynsive.
        """

        test_module = pynsive.import_module('pynsive_test.test_classes')
        self.assertTrue(test_module.SUCCESS)

    def test_listing_classes(self):
        classes = pynsive.list_classes('pynsive_test')
        self.assertEqual(len(classes), 2)

    def test_listing_classes_with_filter(self):
        test_module = pynsive.import_module('pynsive_test.test_classes')

        def subclasses_only(test_type):
            same = test_type is not test_module.PynsiveTestingClass
            is_subclass = issubclass(
                test_type, test_module.PynsiveTestingClass)
            return not same and is_subclass

        classes = pynsive.list_classes('pynsive_test', subclasses_only)
        self.assertEqual(len(classes), 1)

    def test_discovering_modules(self):
        stock_path_packages = [
            'pynsive.plugin.manager',
            'pynsive.plugin.loader']
        plugin_path_packages = ['pynsive_test.test_classes']
        self.assertEqual(
            plugin_path_packages,
            pynsive.discover_modules('pynsive_test'))
        self.assertEqual(
            stock_path_packages,
            pynsive.discover_modules('pynsive.plugin'))

    def test_discovering_classes(self):
        classes = pynsive.discover_classes('pynsive_test')
        self.assertEqual(2, len(classes))

    def test_discovering_classes_with_filter(self):
        test_module = pynsive.import_module('pynsive_test.test_classes')

        def subclasses_only(test_type):
            same = test_type is not test_module.PynsiveTestingClass
            is_subclass = issubclass(
                test_type, test_module.PynsiveTestingClass)
            return not same and is_subclass

        classes = pynsive.discover_classes('pynsive_test', subclasses_only)
        self.assertEqual(len(classes), 1)



import imp
import sys
import os

from ..common import *


class LoaderError(ImportError):
    """
    This error is thrown when the module loader encounters an exception or
    an unrecoverable state while attempting to load a dynamically located
    module.
    """

    def __init__(self, msg):
        self.msg = msg


class ModuleLoader(object):
    """
    As per PEP302, this module loader provides all of the necessarry context
    for loading a python file and executing its contents.
    """

    def __init__(self, module_path, module_name, load_target, is_pkg):
        self.module_path = module_path
        self.module_name = module_name
        self.load_target = load_target
        self.is_pkg = is_pkg

    def _read_code(self):
        """
        Simple abstraction to make reading a module file easy.
        """
        fin = open(self.load_target, 'r')
        code = fin.read()
        fin.close()
        return code

    def load_module(self, fullname):
        """
        Loads a module's code and sets the module's expected hidden
        variables. For more information on these variables and what they
        are for, please see PEP302.
        """
        if fullname != self.module_name:
            raise LoaderError(
                'Requesting a module that the loader is unaware of.')

        if fullname in sys.modules:
            return sys.modules[fullname]

        code = self._read_code()
        module = imp.new_module(fullname)
        module.__file__ = self.load_target
        module.__loader__ = self

        if self.is_pkg:
            module.__path__ = [self.module_path]
            module.__package__ = fullname
        else:
            module.__package__ = fullname.rpartition('.')[0]

        sys.modules[fullname] = module
        exec(code, module.__dict__)
        return module


class ModuleFinder(object):
    """
    As per PEP302, this module loader provides all of the necessarry context
    for dynamically locating python modules based. This finder searches
    directories based on paths added to its internal list of available
    search directories.
    """

    def __init__(self, paths=None):
        self.paths = paths ? paths : list()

    def add_path(self, new_path):
        """
        Adds a path to search through when attempting to look up a module.
        """
        if new_path not in self.paths:
            self.paths.append(new_path)

    def find_module(self, fullname, path=None):
        """
        Searches the paths for the required module.
        """
        module_path = os.path.join(*fullname.split(MODULE_PATH_SEP))

        for plugin_path in self.paths:
            target_path = os.path.join(plugin_path, module_path)
            is_pkg = False

            # If the target references a directory, try to load it as
            # a module by referencing the __init__.py file, otherwise
            # append .py and attempt to resolve it.
            if os.path.isdir(target_path):
                target_file = os.path.join(target_path, '__init__.py')
                is_pkg = True
            else:
                target_file = '{}.py'.format(target_path)

            if os.path.exists(target_file):
                return ModuleLoader(target_path, fullname, target_file, is_pkg)

        return None



import sys
import os.path
import inspect

from .common import *


def _list_classes(module, type_filter):
    found = list()

    for name, module_obj in inspect.getmembers(module):
        if inspect.isclass(module_obj):
            append = not type_filter or type_filter(module_obj)
            if append:
                found.append(module_obj)
    return found


def _should_use_module(module):
    return hasattr(module, PATH_ATTRUBITE) and len(module.__path__) > 0


def _scan_paths_for(mname, paths):
    mpath_part = mname.replace(MODULE_PATH_SEP, os.sep)
    for path in paths:
        scan_target = os.path.join(path, mpath_part)
        is_dir = os.path.isdir(scan_target)
        is_file = os.path.isfile('{}.py'.format(scan_target))
        if is_dir or is_file:
            return mpath


def _scan_dir(directory):
    found = list()
    module_py = os.path.join(directory, MODULE_INIT_FILE)
    if os.path.isfile(module_py):
        # Scan only if there's an __init__.py file
        for dir_entry in os.listdir(directory):
            if dir_entry != MODULE_INIT_FILE and dir_entry.endswith('.py'):
                found.append(dir_entry.rstrip('.py'))
    return found


def list_classes(mname, type_filter=None):
    """
    Attempts to list all of the clases within a specified module. This
    function works for modules located in the default path as well as
    extended paths via the sys.meta_path hooks.

    If a type filter is set, it will be called with each class as its
    parameter. This filter's return value must be interpretable as a
    boolean. Results that evaluate as True will include the type in the
    list of returned classes. Results that evaluate as False will excluse
    the type in the list of returned classes.
    """
    found = list()
    module = import_module(mname)
    if inspect.ismodule(module):
        [found.append(mod) for mod in _list_classes(module, type_filter)]
    return found


def discover_modules(mname):
    """
    Attempts to list all of the submodules under a module. This function
    works for modules located in the default path as well as extended paths
    via the sys.meta_path hooks.

    This funciton carries the expectation that the hidden module variable
    '__path__' has been set correctly.
    """
    found = list()
    module = import_module(mname)

    if module and _should_use_module(module):
        mpath = module.__path__[0]
    else:
        mpaths = sys.path
        mpath = _scan_paths_for(mname, mpaths)

    if mpath:
        for pmname in _scan_dir(mpath):
            found_mod = '{}{}{}'.format(mname, MODULE_PATH_SEP, pmname)
            found.append(found_mod)
    return found


def discover_classes(module, type_filter=None):
    """
    Attempts to list all of the classes within a given module namespace.
    This method, unlike list_classes, will recurse into discovered
    submodules.

    If a type filter is set, it will be called with each class as its
    parameter. This filter's return value must be interpretable as a
    boolean. Results that evaluate as True will include the type in the
    list of returned classes. Results that evaluate as False will excluse
    the type in the list of returned classes.
    """
    classes = list()
    mnames = discover_modules(module)
    for mname in mnames:
        [classes.append(c) for c in list_classes(mname, type_filter)]
    return classes



# -*- coding: utf-8 -*-
try:
    from setuptools import setup, find_packages
    from setuptools.command import easy_install
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages
    from setuptools.command import easy_install

import sys
from os import path
from distutils.core import setup
from distutils.extension import Extension


def read(relative):
    contents = open(relative, 'r').read()
    return [l for l in contents.split('\n') if l != '']


def ez_install(package):
    easy_install.main(["-U", package])


setup(
    name='pynsive',
    version='0.1.2',
    description=("""A Python plugin and introspection library prounounced,
"pensive." This library allows a developer to introduce plugins via PEP 302
and allows for class searching via type and inheritance introspections.
"""),
    author='John Hopper',
    author_email='john.hopper@jpserver.net',
    url='https://github.com/zinic/pynsive',
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Environment :: Plugins',
        'Programming Language :: Python',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3'
    ],
    tests_require=read('./tools/test-requires'),
    install_requires=read('./tools/install-requires'),
    test_suite='nose.collector',
    zip_safe=False,
    include_package_data=True,
    packages=find_packages(exclude=['ez_setup']))
